/* editor.c
 *
 * Copyright 2002-2003 Vesa Halttunen
 *
 * This file is part of Tutka.
 *
 * Tutka is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Tutka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tutka; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include "song.h"
#include "editor.h"
#include "player.h"

/* Initializes the editor */
struct editor *editor_open(char *guifilename) {
  struct editor *editor=(struct editor *)calloc(1, sizeof(struct editor));

  editor->space=1;

  /* Open a GUI file */
  editor->gui=gui_open(editor, guifilename);

  return editor;
}

/* Opens an existing song or creates new */
void editor_song_open(struct editor *editor, char *filename) {
  if(editor==NULL)
    return;

  /* Close the old player */
  if(editor->player!=NULL)
    player_close(editor->player);

  /* Free the old song */
  if(editor->song!=NULL)
    song_free(editor->song);

  if(editor->filename!=NULL)
    free(editor->filename);
    
  /* If a filename was supplied open the file */
  if(filename!=NULL) {
    editor->filename=strdup(filename);
    editor->song=song_load(filename);
  } else {
    editor->filename=NULL;
    editor->song=song_alloc();
  }

  /* Open the player */
  editor->player=player_open(editor->song, editor);

  /* Refresh the GUI */
  gui_refresh_all(editor->gui);
}

/* Deinitializes the editor */
void editor_close(struct editor *editor) {
  if(editor==NULL)
    return;

  /* Close the GUI */
  if(editor->gui!=NULL)
    gui_close(editor->gui);

  /* Close the player */
  if(editor->player!=NULL)
    player_close(editor->player);

  /* Free the song */
  if(editor->song!=NULL)
    song_free(editor->song);

  if(editor->filename!=NULL)
    free(editor->filename);

  /* Free the copy & paste buffers */
  if(editor->copyblock!=NULL)
    block_free(editor->copyblock);

  if(editor->copytrack!=NULL)
    block_free(editor->copytrack);

  if(editor->copyarea!=NULL)
    block_free(editor->copyarea);

  free(editor);
}

/* Changes the position of the editor and refreshes the GUI accordingly */
void editor_set_values(struct editor *editor,
		       int section, int playseq, int position, int block,
		       int line, int time, int tempochanged) {
  if(section!=editor->section &&
     section>=0 && section<editor->song->numsections) {
    gui_set_refresh(editor->gui, GUI_REFRESH_INFO|GUI_REFRESH_SECTIONLIST);
    editor->section=section;
  }
  if(playseq!=editor->playseq &&
     playseq>=0 && playseq<editor->song->numplayseqs) {
    gui_set_refresh(editor->gui, GUI_REFRESH_INFO|GUI_REFRESH_PLAYSEQ);
    editor->playseq=playseq;
  }
  if(position!=editor->position &&
     position>=0 && position<editor->song->playseqs[editor->playseq]->length) {
    gui_set_refresh(editor->gui, GUI_REFRESH_INFO|GUI_REFRESH_PLAYSEQ);
    editor->position=position;
  }
  if(block!=editor->block &&
     block>=0 && block<editor->song->numblocks) {
    gui_set_refresh(editor->gui,
		    GUI_REFRESH_INFO|
		    GUI_REFRESH_BLOCKLIST|GUI_REFRESH_TRACKER);
    editor->block=block;
  }
  if(line!=editor->line &&
     line>=0 && line<editor->song->blocks[editor->block]->length) {
    gui_set_refresh(editor->gui, GUI_REFRESH_TRACKER);
    editor->line=line;
  }
  if(time!=editor->time) {
    gui_set_refresh(editor->gui, GUI_REFRESH_TIMER);
    editor->time=time;
  }
  if(tempochanged)
    gui_set_refresh(editor->gui, GUI_REFRESH_TEMPO);
}

/* Changes the position of the editor and refreshes the GUI accordingly */
void editor_set_section_immediate(struct editor *editor, int section) {
  if(section!=editor->section) {
    if(section<0)
      section=0;

    if(section>=editor->song->numsections)
      section=editor->song->numsections-1;
      
    editor->section=section;

    gui_info_refresh(editor->gui); 
    gui_sectionlist_refresh(editor->gui, FALSE);
  }
}

/* Changes the position of the editor and refreshes the GUI accordingly */
void editor_set_playseq_immediate(struct editor *editor, int playseq) {
  if(playseq!=editor->playseq) {
    if(playseq<0)
      playseq=0;

    if(playseq>=editor->song->numplayseqs)
      playseq=editor->song->numplayseqs-1;

    editor->playseq=playseq;

    gui_info_refresh(editor->gui);
    gui_playseq_refresh(editor->gui, FALSE);
  }
}

/* Changes the position of the editor and refreshes the GUI accordingly */
void editor_set_position_immediate(struct editor *editor, int position) {
  if(position!=editor->position) {
    if(position<0)
      position=0;

    if(position>=editor->song->playseqs[editor->playseq]->length)
      position=editor->song->playseqs[editor->playseq]->length-1;

    editor->position=position;

    gui_info_refresh(editor->gui);
    gui_playseq_refresh(editor->gui, FALSE);
  }
}

/* Changes the position of the editor and refreshes the GUI accordingly */
void editor_set_block_immediate(struct editor *editor, int block) {
  if(block!=editor->block) {
    if(block<0)
      block=0;

    if(block>=editor->song->numblocks)
      block=editor->song->numblocks-1;

    if(editor->cmdpage>=editor->song->blocks[block]->commandpages) {
      while(editor->cmdpage>=editor->song->blocks[block]->commandpages)
	editor->cmdpage--;
    }

    editor->block=block;

    gui_info_refresh(editor->gui);
    gui_blocklist_refresh(editor->gui, FALSE);
    gui_tracker_refresh(editor->gui);
  }
}

/* Changes the position of the editor and refreshes the GUI accordingly */
void editor_set_line_immediate(struct editor *editor, int line) {
  if(line!=editor->line) {
    while(line<0)
      line+=editor->song->blocks[editor->block]->length;

    while(line>=editor->song->blocks[editor->block]->length)
      line-=editor->song->blocks[editor->block]->length;
    
    editor->line=line;

    gui_tracker_refresh(editor->gui);
  }
}

/* Changes the position of the editor and refreshes the GUI accordingly */
void editor_set_commandpage_immediate(struct editor *editor, int cmdpage) {
  if(cmdpage!=editor->cmdpage) {
    while(cmdpage<0)
      cmdpage+=editor->song->blocks[editor->block]->commandpages;

    while(cmdpage>=editor->song->blocks[editor->block]->commandpages)
      cmdpage-=editor->song->blocks[editor->block]->commandpages;

    editor->cmdpage=cmdpage;
    
    gui_info_refresh(editor->gui);
    gui_tracker_refresh(editor->gui);
  }
}

/* Refreshes playseq from section and block from position */
void editor_refresh_playseq_and_block(struct editor *editor) {
  struct song *song=editor->song;

  if(editor->section>=song->numsections)
    editor->section=0;

  editor->playseq=song->sections[editor->section];

  if(editor->position>=song->playseqs[editor->playseq]->length) {
    editor->position=0;
    editor_refresh_playseq_and_block(editor);
  } else
    editor->block=song->playseqs[editor->playseq]->blocknumbers[editor->position];
}
